"""
Defines the abstract base traits of a tox environment.
"""
import logging
import os
import re
import shutil
import sys
from abc import ABC, abstractmethod
from contextlib import contextmanager
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Sequence, Tuple, Union, cast

from tox.config.main import Config
from tox.config.sets import ConfigSet
from tox.execute.api import Execute, ExecuteStatus, Outcome, StdinSource
from tox.execute.request import ExecuteRequest
from tox.journal import EnvJournal
from tox.report import OutErr, ToxHandler
from tox.tox_env.errors import Recreate

from .info import Info

if TYPE_CHECKING:
    from tox.config.cli.parser import Parsed

LOGGER = logging.getLogger(__name__)


class ToxEnv(ABC):
    def __init__(
        self, conf: ConfigSet, core: ConfigSet, options: "Parsed", journal: EnvJournal, log_handler: ToxHandler
    ) -> None:
        self.journal = journal
        self.conf: ConfigSet = conf
        self.core: ConfigSet = core
        self.options = options
        self._executor: Optional[Execute] = None
        self.register_config()
        self._cache = Info(self.conf["env_dir"])
        self._paths: List[Path] = []
        self._hidden_outcomes: Optional[List[Outcome]] = []
        self.log_handler = log_handler
        self._env_vars: Optional[Dict[str, str]] = None
        self._suspended_out_err: Optional[OutErr] = None
        self.setup_done = False
        self.clean_done = False
        self._execute_statuses: Dict[int, ExecuteStatus] = {}
        self._interrupted = False

    def interrupt(self) -> None:
        logging.warning("interrupt tox environment: %s", self.conf.name)
        self._interrupted = True
        for status in list(self._execute_statuses.values()):
            status.interrupt()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.conf['env_name']})"

    @property
    def executor(self) -> Execute:
        if self._executor is None:
            self._executor = self.build_executor()
        return self._executor

    @property
    def has_display_suspended(self) -> bool:
        return self._suspended_out_err is not None

    @abstractmethod
    def build_executor(self) -> Execute:
        raise NotImplementedError

    def register_config(self) -> None:
        self.conf.add_constant(
            keys=["env_name", "envname"],
            desc="the name of the tox environment",
            value=self.conf.name,
        )
        self.conf.add_config(
            keys=["env_dir", "envdir"],
            of_type=Path,
            default=lambda conf, name: cast(Path, conf.core["work_dir"]) / cast(str, self.conf["env_name"]),
            desc="directory assigned to the tox environment",
        )
        self.conf.add_config(
            keys=["env_tmp_dir", "envtmpdir"],
            of_type=Path,
            default=lambda conf, name: cast(Path, conf.core["work_dir"]) / cast(str, self.conf["env_name"]) / "tmp",
            desc="a folder that is always reset at the start of the run",
        )

        def set_env_post_process(values: Dict[str, str], config: Config) -> Dict[str, str]:
            env = self.default_set_env()
            env.update(values)
            return env

        self.conf.add_config(
            keys=["set_env", "setenv"],
            of_type=Dict[str, str],
            default={},
            desc="environment variables to set when running commands in the tox environment",
            post_process=set_env_post_process,
        )

        def pass_env_post_process(values: List[str], config: Config) -> List[str]:
            values.extend(self.default_pass_env())
            return sorted(list({k: None for k in values}.keys()))

        self.conf.add_config(
            keys=["pass_env", "passenv"],
            of_type=List[str],
            default=[],
            desc="environment variables to pass on to the tox environment",
            post_process=pass_env_post_process,
        )

        self.conf.add_config(
            "parallel_show_output",
            of_type=bool,
            default=False,
            desc="if set to True the content of the output will always be shown  when running in parallel mode",
        )

    def default_set_env(self) -> Dict[str, str]:
        return {}

    def default_pass_env(self) -> List[str]:
        env = [
            "https_proxy",
            "http_proxy",
            "no_proxy",
        ]
        if sys.stdout.isatty():  # if we're on a interactive shell pass on the TERM
            env.append("TERM")
        if sys.platform == "win32":  # pragma: win32 cover
            env.extend(
                [
                    "TEMP",
                    "TMP",
                ]
            )
        else:  # pragma: win32 no cover
            env.append("TMPDIR")
        return env

    def setup(self) -> None:
        """
        1. env dir exists
        2. contains a runner with the same type.
        """
        conf = {"name": self.conf.name, "type": type(self).__name__}
        try:
            with self._cache.compare(conf, ToxEnv.__name__) as (eq, old):
                if eq is False and old is not None:  # recreate if already created and not equals
                    logging.warning(f"env type changed from {old} to {conf}, will recreate")
                    raise Recreate  # recreate if already exists and type changed
                self.setup_done, self.clean_done = True, False
        finally:
            self._handle_env_tmp_dir()

    def ensure_setup(self, recreate: bool = False) -> None:
        if self.setup_done is True:
            return
        if recreate:
            self.clean()
        try:
            self.setup()
        except Recreate:
            if not recreate:  # pragma: no cover
                self.clean(force=True)
                self.setup()
        self.setup_has_been_done()

    def setup_has_been_done(self) -> None:
        """called when setup is done"""

    def _handle_env_tmp_dir(self) -> None:
        """Ensure exists and empty"""
        env_tmp_dir: Path = self.conf["env_tmp_dir"]
        if env_tmp_dir.exists() and next(env_tmp_dir.iterdir(), None) is not None:
            LOGGER.debug("clear env temp folder %s", env_tmp_dir)
            shutil.rmtree(env_tmp_dir, ignore_errors=True)
        env_tmp_dir.mkdir(parents=True, exist_ok=True)

    def clean(self, force: bool = False) -> None:
        if self.clean_done:  # pragma: no branch
            return  # pragma: no cover
        env_dir: Path = self.conf["env_dir"]
        if env_dir.exists():
            LOGGER.warning("remove tox env folder %s", env_dir)
            shutil.rmtree(env_dir)
        self._cache.reset()
        self.setup_done, self.clean_done = False, True

    @property
    def environment_variables(self) -> Dict[str, str]:
        if self._env_vars is not None:
            return self._env_vars
        result: Dict[str, str] = {}

        pass_env: List[str] = self.conf["pass_env"]
        glob_pass_env = [re.compile(e.replace("*", ".*")) for e in pass_env if "*" in e]
        literal_pass_env = [e for e in pass_env if "*" not in e]
        for env in literal_pass_env:
            if env in os.environ:  # pragma: no branch
                result[env] = os.environ[env]
        if glob_pass_env:  # pragma: no branch
            for env, value in os.environ.items():
                if any(g.match(env) is not None for g in glob_pass_env):
                    result[env] = value
        set_env: Dict[str, str] = self.conf["set_env"]
        result.update(set_env)
        result["PATH"] = os.pathsep.join([str(i) for i in self._paths] + os.environ.get("PATH", "").split(os.pathsep))
        self._env_vars = result
        return result

    def execute(
        self,
        cmd: Sequence[Union[Path, str]],
        stdin: StdinSource,
        show: Optional[bool] = None,
        cwd: Optional[Path] = None,
        run_id: str = "",
        executor: Optional[Execute] = None,
    ) -> Outcome:
        with self.execute_async(cmd, stdin, show, cwd, run_id, executor) as status:
            while status.exit_code is None:
                status.wait()
        if status.outcome is None:  # pragma: no cover # this should not happen
            raise RuntimeError  # pragma: no cover
        return status.outcome

    @contextmanager
    def execute_async(
        self,
        cmd: Sequence[Union[Path, str]],
        stdin: StdinSource,
        show: Optional[bool] = None,
        cwd: Optional[Path] = None,
        run_id: str = "",
        executor: Optional[Execute] = None,
    ) -> Iterator[ExecuteStatus]:
        if self._interrupted:
            raise SystemExit(-2)
        if cwd is None:
            cwd = self.core["tox_root"]
        if show is None:
            show = self.options.verbosity > 3
        request = ExecuteRequest(cmd, cwd, self.environment_variables, stdin, run_id)
        if _CWD == request.cwd:
            repr_cwd = ""
        else:
            try:
                repr_cwd = f" {_CWD.relative_to(cwd)}"
            except ValueError:
                repr_cwd = f" {cwd}"
        LOGGER.warning("%s%s> %s", run_id, repr_cwd, request.shell_cmd)
        out_err = self.log_handler.stdout, self.log_handler.stderr
        if executor is None:
            executor = self.executor
        with self._execute_call(executor, out_err, request, show) as execute_status:
            execute_id = id(execute_status)
            try:
                self._execute_statuses[execute_id] = execute_status
                yield execute_status
            finally:
                self._execute_statuses.pop(execute_id)
        if show and self._hidden_outcomes is not None:
            if execute_status.outcome is not None:  # pragma: no cover # if it gets cancelled before even starting
                self._hidden_outcomes.append(execute_status.outcome)
        if self.journal and execute_status.outcome is not None:
            self.journal.add_execute(execute_status.outcome, run_id)

    @staticmethod
    @contextmanager
    def _execute_call(
        executor: Execute, out_err: OutErr, request: ExecuteRequest, show: bool
    ) -> Iterator[ExecuteStatus]:
        with executor.call(
            request=request,
            show=show,
            out_err=out_err,
        ) as execute_status:
            yield execute_status

    @staticmethod
    @abstractmethod
    def id() -> str:
        raise NotImplementedError

    @contextmanager
    def display_context(self, suspend: bool) -> Iterator[None]:
        with self.log_context():
            with self.log_handler.suspend_out_err(suspend, self._suspended_out_err) as out_err:
                if suspend:  # only set if suspended
                    self._suspended_out_err = out_err
                yield

    def close_and_read_out_err(self) -> Optional[Tuple[bytes, bytes]]:
        if self._suspended_out_err is None:  # pragma: no branch
            return None  # pragma: no cover
        (out, err), self._suspended_out_err = self._suspended_out_err, None
        out_b, err_b = cast(BytesIO, out.buffer).getvalue(), cast(BytesIO, err.buffer).getvalue()
        out.close()
        err.close()
        return out_b, err_b

    @contextmanager
    def log_context(self) -> Iterator[None]:
        with self.log_handler.with_context(cast(str, self.conf.name)):
            yield

    def teardown(self) -> None:
        """Any cleanup operation on environment done"""


_CWD = Path.cwd()



"""Helper methods related to graph theory."""
from collections import OrderedDict, defaultdict
from typing import Dict, List, Optional, Set


def stable_topological_sort(graph: Dict[str, Set[str]]) -> List[str]:
    to_order = set(graph.keys())  # keep a log of what  we need to order

    # normalize graph - fill missing nodes (assume no dependency)
    for values in list(graph.values()):
        for value in values:
            if value not in graph:
                graph[value] = set()

    inverse_graph = defaultdict(set)
    for key, depends in graph.items():
        for depend in depends:
            inverse_graph[depend].add(key)

    topology = []
    degree = {k: len(v) for k, v in graph.items()}
    ready_to_visit = {n for n, d in degree.items() if not d}
    need_to_visit = OrderedDict((i, None) for i in graph.keys())
    while need_to_visit:
        # to keep stable, pick the first node ready to visit in the original order
        for node in need_to_visit:
            if node in ready_to_visit:
                break
        else:
            break
        del need_to_visit[node]

        topology.append(node)

        # decrease degree for nodes we're going too
        for to_node in inverse_graph[node]:
            degree[to_node] -= 1
            if not degree[to_node]:  # if a node has no more incoming node it's ready to visit
                ready_to_visit.add(to_node)

    result = [n for n in topology if n in to_order]  # filter out missing nodes we extended

    if len(result) < len(to_order):
        identify_cycle(graph)
        msg = "could not order tox environments and failed to detect circle"  # pragma: no cover
        raise ValueError(msg)  # pragma: no cover
    return result


def identify_cycle(graph: Dict[str, Set[str]]) -> None:
    path: Dict[str, None] = OrderedDict()
    visited = set()

    def visit(vertex: str) -> Optional[Dict[str, None]]:
        if vertex in visited:
            return None
        visited.add(vertex)
        path[vertex] = None
        for neighbour in graph.get(vertex, ()):
            if neighbour in path or visit(neighbour):
                return path
        del path[vertex]
        return None

    for node in graph:  # pragma: no branch # we never get here if the graph is empty
        result = visit(node)
        if result is not None:
            raise ValueError(f"{' | '.join(result.keys())}")



import sys
from itertools import zip_longest
from textwrap import dedent

import pytest
from packaging.requirements import Requirement

from tox.pytest import TempPathFactory, ToxProjectCreator
from tox.tox_env.python.virtual_env.package.api import PackageType, Pep517VirtualEnvPackage
from tox.util.pep517.via_fresh_subprocess import SubprocessFrontend

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from importlib.metadata import Distribution, PathDistribution  # type: ignore[attr-defined]
else:  # pragma: no cover (<py38)
    from importlib_metadata import Distribution, PathDistribution  # noqa


@pytest.mark.parametrize(
    "pkg_type, of_type",
    [
        ("dev", "virtualenv-legacy-dev"),
        ("sdist", "virtualenv-pep-517-sdist"),
        ("wheel", "virtualenv-pep-517-wheel"),
    ],
)
def test_tox_ini_package_type_valid(tox_project: ToxProjectCreator, pkg_type: str, of_type: str) -> None:
    proj = tox_project({"tox.ini": f"[testenv]\npackage={pkg_type}"})
    result = proj.run("c", "-k", "package_tox_env_type")
    result.assert_success()
    res = result.state.tox_env("py").conf["package"]
    assert res is getattr(PackageType, pkg_type)
    got_type = result.state.tox_env("py").conf["package_tox_env_type"]
    assert got_type == of_type


def test_tox_ini_package_type_invalid(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({"tox.ini": "[testenv]\npackage=bad"})
    result = proj.run("c", "-k", "package_tox_env_type")
    result.assert_failed()
    assert " invalid package config type 'bad' requested, must be one of sdist, wheel, dev, skip" in result.out


@pytest.fixture(scope="session")
def pkg_with_extras(tmp_path_factory: TempPathFactory) -> PathDistribution:  # type: ignore[no-any-unimported]
    py_ver = ".".join(str(i) for i in sys.version_info[0:2])
    setup_cfg = f"""
    [metadata]
    name = demo
    [options]
    packages = find:
    install_requires =
        appdirs>=1.4.3
        colorama>=0.4.3

    [options.extras_require]
    testing =
        covdefaults>=1.2; python_version == '2.7' or python_version == '{py_ver}'
        pytest>=5.4.1; python_version == '{py_ver}'
    docs =
        sphinx>=3
        sphinx-rtd-theme>=0.4.3,<1
    format =
        black>=3
        flake8
    """
    tmp_path = tmp_path_factory.mktemp("prj")
    (tmp_path / "setup.cfg").write_text(dedent(setup_cfg))
    (tmp_path / "setup.py").write_text("from setuptools import setup; setup()")
    toml = '[build-system]\nrequires=["setuptools", "wheel"]\nbuild-backend = "setuptools.build_meta"'
    (tmp_path / "pyproject.toml").write_text(toml)
    frontend = SubprocessFrontend(*SubprocessFrontend.create_args_from_folder(tmp_path)[:-1])
    meta = tmp_path / "meta"
    result = frontend.prepare_metadata_for_build_wheel(meta)
    return Distribution.at(result.metadata)


def test_load_dependency_no_extra(pkg_with_extras: PathDistribution) -> None:  # type: ignore[no-any-unimported]
    result = Pep517VirtualEnvPackage.discover_package_dependencies(pkg_with_extras, set())
    for left, right in zip_longest(result, (Requirement("appdirs>=1.4.3"), Requirement("colorama>=0.4.3"))):
        assert isinstance(right, Requirement)
        assert str(left) == str(right)


def test_load_dependency_many_extra(pkg_with_extras: PathDistribution) -> None:  # type: ignore[no-any-unimported]
    py_ver = ".".join(str(i) for i in sys.version_info[0:2])
    result = Pep517VirtualEnvPackage.discover_package_dependencies(pkg_with_extras, {"docs", "testing"})
    exp = [
        Requirement("appdirs>=1.4.3"),
        Requirement("colorama>=0.4.3"),
        Requirement("sphinx>=3"),
        Requirement("sphinx-rtd-theme<1,>=0.4.3"),
        Requirement(f'covdefaults>=1.2; python_version == "2.7" or python_version == "{py_ver}"'),
        Requirement(f'pytest>=5.4.1; python_version == "{py_ver}"'),
    ]
    for left, right in zip_longest(result, exp):
        assert isinstance(right, Requirement)
        assert str(left) == str(right)



"""Handles communication on the backend side between frontend and backend"""
from typing import Any, Optional, Sequence

class MissingCommand(TypeError): ...

class BackendProxy:
    backend_module: str
    backend_object: Optional[str]
    backend: Any
    def __init__(self, backend_module: str, backend_obj: Optional[str]) -> None: ...
    def __call__(self, name: str, *args: Any, **kwargs: Any) -> Any: ...
    def __str__(self) -> str: ...
    def _exit(self) -> None: ...

def run(argv: Sequence[str]) -> int: ...
def flush() -> None: ...



import pytest
from pytest_mock import MockerFixture

from tox.pytest import ToxProjectCreator
from tox.report import HandledError
from tox.run import run


@pytest.mark.parametrize("exception", [HandledError, KeyboardInterrupt])
def test_exit_code_minus_2_on_expected_exit(exception: Exception, mocker: MockerFixture) -> None:
    mocker.patch("tox.run.main", side_effect=exception)
    with pytest.raises(SystemExit) as system_exit:
        run()
    assert system_exit.value.code == -2


def test_re_raises_on_unexpected_exit(mocker: MockerFixture) -> None:
    mocker.patch("tox.run.main", side_effect=ValueError)
    with pytest.raises(ValueError):
        run()


def test_no_tox_ini(tox_project: ToxProjectCreator) -> None:
    project = tox_project({})
    outcome = project.run("l")
    # assume an empty tox.ini at the cwd level
    assert outcome.state.options.work_dir is None


def test_custom_work_dir(tox_project: ToxProjectCreator) -> None:
    project = tox_project({})
    outcome = project.run("c", "--workdir", str(project.path.parent))
    assert outcome.state.options.work_dir == project.path.parent



"""
Run tox environments in sequential order.
"""
from tox.config.cli.parser import ToxParser
from tox.plugin.impl import impl
from tox.session.common import env_list_flag
from tox.session.state import State

from .common import env_run_create_flags, execute


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command("run", ["r"], "run environments", run_sequential)
    env_list_flag(our)
    env_run_create_flags(our)


def run_sequential(state: State) -> int:
    return execute(state, max_workers=1, spinner=False, live=True)



from argparse import ArgumentParser, ArgumentTypeError, Namespace
from typing import Optional

import pytest

from tox.session.cmd.run.common import SkipMissingInterpreterAction


@pytest.mark.parametrize("values", ["config", None, "true", "false"])
def test_skip_missing_interpreter_action_ok(values: Optional[str]) -> None:
    args_namespace = Namespace()
    SkipMissingInterpreterAction(option_strings=["-i"], dest="into")(ArgumentParser(), args_namespace, values)
    expected = "true" if values is None else values
    assert args_namespace.into == expected


def test_skip_missing_interpreter_action_nok() -> None:
    argument_parser = ArgumentParser()
    with pytest.raises(ArgumentTypeError, match=r"value must be 'config', 'true', or 'false' \(got 'bad value'\)"):
        SkipMissingInterpreterAction(option_strings=["-i"], dest="into")(argument_parser, Namespace(), "bad value")



import multiprocessing

from pytest_mock import MockerFixture

from tox.util.cpu import auto_detect_cpus


def test_auto_detect_cpus() -> None:
    num_cpus_actual = multiprocessing.cpu_count()
    assert auto_detect_cpus() == num_cpus_actual


def test_auto_detect_cpus_returns_one_when_cpu_count_throws_NotImplementedError(mocker: MockerFixture) -> None:
    mocker.patch.object(multiprocessing, "cpu_count", side_effect=NotImplementedError)
    assert auto_detect_cpus() == 1



import sys
from abc import ABC, abstractmethod
from collections import OrderedDict
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Generic, Iterator, List, Set, Tuple, Type, TypeVar, Union, cast

from ..types import Command, EnvList

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from typing import Literal
else:  # pragma: no cover (py38+)
    from typing_extensions import Literal  # noqa


_NO_MAPPING = object()
T = TypeVar("T")
V = TypeVar("V")


class Convert(ABC, Generic[T]):
    """A class that converts a raw type to a given tox (python) type"""

    def to(self, raw: T, of_type: Type[V]) -> V:
        from_module = getattr(of_type, "__module__", None)
        if from_module in ("typing", "typing_extensions"):
            return self._to_typing(raw, of_type)  # type: ignore[return-value]
        elif issubclass(of_type, Path):
            return self.to_path(raw)  # type: ignore[return-value]
        elif issubclass(of_type, bool):
            return self.to_bool(raw)  # type: ignore[return-value]
        elif issubclass(of_type, Command):
            return self.to_command(raw)  # type: ignore[return-value]
        elif issubclass(of_type, EnvList):
            return self.to_env_list(raw)  # type: ignore[return-value]
        elif issubclass(of_type, str):
            return self.to_str(raw)  # type: ignore[return-value]
        elif isinstance(raw, of_type):
            return raw
        elif issubclass(of_type, Enum):
            return cast(V, getattr(of_type, str(raw)))
        return of_type(raw)  # type: ignore[call-arg]

    def _to_typing(self, raw: T, of_type: Type[V]) -> V:
        origin = getattr(of_type, "__origin__", of_type.__class__)
        result: Any = _NO_MAPPING
        if origin in (list, List):
            entry_type = of_type.__args__[0]  # type: ignore[attr-defined]
            result = [self.to(i, entry_type) for i in self.to_list(raw, entry_type)]
        elif origin in (set, Set):
            entry_type = of_type.__args__[0]  # type: ignore[attr-defined]
            result = {self.to(i, entry_type) for i in self.to_set(raw, entry_type)}
        elif origin in (dict, Dict):
            key_type, value_type = of_type.__args__[0], of_type.__args__[1]  # type: ignore[attr-defined]
            result = OrderedDict(
                (self.to(k, key_type), self.to(v, value_type)) for k, v in self.to_dict(raw, (key_type, value_type))
            )
        elif origin == Union:  # handle Optional values
            args: List[Type[Any]] = of_type.__args__  # type: ignore[attr-defined]
            none = type(None)
            if len(args) == 2 and none in args:
                if isinstance(raw, str):
                    raw = raw.strip()  # type: ignore[assignment]
                if not raw:
                    result = None
                else:
                    new_type = next(i for i in args if i != none)  # pragma: no cover # this will always find a element
                    result = self.to(raw, new_type)
        elif origin == Literal or origin == type(Literal):
            if sys.version_info >= (3, 7):  # pragma: no cover (py37+)
                choice = of_type.__args__
            else:  # pragma: no cover (py38+)
                choice = of_type.__values__  # type: ignore[attr-defined]
            if raw not in choice:
                raise ValueError(f"{raw} must be one of {choice}")
            result = raw
        if result is not _NO_MAPPING:
            return cast(V, result)
        raise TypeError(f"{raw} cannot cast to {of_type!r}")

    @staticmethod
    @abstractmethod
    def to_str(value: T) -> str:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_list(value: T, of_type: Type[Any]) -> Iterator[T]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_set(value: T, of_type: Type[Any]) -> Iterator[T]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_dict(value: T, of_type: Tuple[Type[Any], Type[Any]]) -> Iterator[Tuple[T, T]]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_path(value: T) -> Path:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_command(value: T) -> Command:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_env_list(value: T) -> EnvList:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_bool(value: T) -> bool:
        raise NotImplementedError



"""
Print available tox environments.
"""
from tox.config.cli.parser import ToxParser
from tox.plugin.impl import impl
from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command("list", ["l"], "list environments", list_env)
    our.add_argument("-d", action="store_true", help="list just default envs", dest="list_default_only")
    our.add_argument("--no-desc", action="store_true", help="do not show description", dest="list_no_description")


def list_env(state: State) -> int:
    core = state.conf.core
    option = state.options

    default = core["env_list"]  # this should be something not affected by env-vars :-|
    ignore = {core["provision_tox_env"]}.union(default)

    extra = [] if option.list_default_only else [e for e in state.env_list(everything=True) if e not in ignore]

    if not option.list_no_description and default:
        print("default environments:")
    max_length = max(len(env) for env in (default.envs + extra))

    def report_env(name: str) -> None:
        if not option.list_no_description:
            text = state.tox_env(name).conf["description"]
            if not text.strip():
                text = "[no description]"
            text = text.replace("\n", " ")
            msg = f"{e.ljust(max_length)} -> {text}".strip()
        else:
            msg = e
        print(msg)

    for e in default:
        report_env(e)
    if not option.list_default_only and extra:
        if not option.list_no_description:
            if default:
                print("")
            print("additional environments:")
        for e in extra:
            report_env(e)
    return 0



"""Defines tox error types"""


class Recreate(RuntimeError):
    """Recreate the tox environment"""


class Skip(RuntimeError):
    """Skip this tox environment"""


class Fail(RuntimeError):
    """Failed creating env"""



"""
Abstract base API for executing commands within tox environments.
"""
import logging
import sys
import time
from abc import ABC, abstractmethod
from contextlib import contextmanager
from types import TracebackType
from typing import Callable, Iterator, NoReturn, Optional, Sequence, Tuple, Type

from colorama import Fore

from tox.report import OutErr

from .request import ExecuteRequest, StdinSource
from .stream import SyncWrite

ContentHandler = Callable[[bytes], None]
Executor = Callable[[ExecuteRequest, ContentHandler, ContentHandler], int]
LOGGER = logging.getLogger(__name__)


class ExecuteStatus(ABC):
    def __init__(self, out: SyncWrite, err: SyncWrite) -> None:
        self.outcome: Optional[Outcome] = None
        self._out = out
        self._err = err

    @property
    @abstractmethod
    def exit_code(self) -> Optional[int]:
        raise NotImplementedError

    @abstractmethod
    def wait(self, timeout: Optional[float] = None) -> None:
        raise NotImplementedError

    @abstractmethod
    def write_stdin(self, content: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def interrupt(self) -> None:
        raise NotImplementedError

    def set_out_err(self, out: SyncWrite, err: SyncWrite) -> Tuple[SyncWrite, SyncWrite]:
        res = self._out, self._err
        self._out, self._err = out, err
        return res

    @property
    def out(self) -> bytearray:
        return self._out.content

    @property
    def err(self) -> bytearray:
        return self._err.content


class Execute(ABC):
    """Abstract API for execution of a tox environment"""

    def __init__(self, colored: bool) -> None:
        self._colored = colored

    @contextmanager
    def call(self, request: ExecuteRequest, show: bool, out_err: OutErr) -> Iterator[ExecuteStatus]:
        start = time.monotonic()
        try:
            # collector is what forwards the content from the file streams to the standard streams
            out, err = out_err[0].buffer, out_err[1].buffer
            out_sync = SyncWrite(out.name, out if show else None)
            err_sync = SyncWrite(err.name, err if show else None, Fore.RED if self._colored else None)
            with out_sync, err_sync:
                instance = self.build_instance(request, out_sync, err_sync)
                with instance as status:
                    yield status
                exit_code = status.exit_code
        finally:
            end = time.monotonic()
        status.outcome = Outcome(request, show, exit_code, out_sync.text, err_sync.text, start, end, instance.cmd)

    @abstractmethod
    def build_instance(self, request: ExecuteRequest, out: SyncWrite, err: SyncWrite) -> "ExecuteInstance":
        raise NotImplementedError


class ExecuteInstance(ABC):
    """An instance of a command execution"""

    def __init__(self, request: ExecuteRequest, out: SyncWrite, err: SyncWrite) -> None:
        self.request = request
        self._out = out
        self._err = err

    @property
    def out_handler(self) -> ContentHandler:
        return self._out.handler

    @property
    def err_handler(self) -> ContentHandler:
        return self._err.handler

    @abstractmethod
    def __enter__(self) -> ExecuteStatus:
        raise NotImplementedError

    @abstractmethod
    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def cmd(self) -> Sequence[str]:
        raise NotImplementedError


class Outcome:
    """Result of a command execution"""

    OK = 0

    def __init__(
        self,
        request: ExecuteRequest,
        show_on_standard: bool,
        exit_code: Optional[int],
        out: str,
        err: str,
        start: float,
        end: float,
        cmd: Sequence[str],
    ):
        self.request = request
        self.show_on_standard = show_on_standard
        self.exit_code = exit_code
        self.out = out
        self.err = err
        self.start = start
        self.end = end
        self.cmd = cmd

    def __bool__(self) -> bool:
        return self.exit_code == self.OK

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}: exit {self.exit_code} in {self.elapsed:.2f} seconds"
            f" for {self.request.shell_cmd}"
        )

    def assert_success(self) -> None:
        if self.exit_code is not None and self.exit_code != self.OK:
            self._assert_fail()
        self.log_run_done(logging.INFO)

    def _assert_fail(self) -> NoReturn:
        if self.show_on_standard is False:
            if self.out:
                sys.stdout.write(self.out)
                if not self.out.endswith("\n"):
                    sys.stdout.write("\n")
            if self.err:
                sys.stderr.write(Fore.RED)
                sys.stderr.write(self.err)
                sys.stderr.write(Fore.RESET)
                if not self.err.endswith("\n"):
                    sys.stderr.write("\n")
        self.log_run_done(logging.CRITICAL)
        raise SystemExit(self.exit_code)

    def log_run_done(self, lvl: int) -> None:
        req = self.request
        LOGGER.log(lvl, "exit %s (%.2f seconds) %s> %s", self.exit_code, self.elapsed, req.cwd, req.shell_cmd)

    @property
    def elapsed(self) -> float:
        return self.end - self.start

    def out_err(self) -> Tuple[str, str]:
        return self.out, self.err


__all__ = (
    "ContentHandler",
    "Outcome",
    "Execute",
    "ExecuteInstance",
    "ExecuteStatus",
    "StdinSource",
)



import sys
from configparser import ConfigParser
from pathlib import Path
from typing import Callable, List, Optional

import pytest

from tox.config.main import Config
from tox.config.source.tox_ini import ToxIni

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from typing import Protocol
else:  # pragma: no cover (<py38)
    from typing_extensions import Protocol  # noqa


class ReplaceOne(Protocol):
    def __call__(self, conf: str, pos_args: Optional[List[str]] = None) -> str:
        ...


@pytest.fixture
def replace_one(mk_ini_conf: Callable[[str], ConfigParser], tmp_path: Path) -> ReplaceOne:
    def example(conf: str, pos_args: Optional[List[str]] = None) -> str:
        tox_ini_file = tmp_path / "tox.ini"
        tox_ini_file.write_text(f"[testenv:py]\nenv={conf}\n")
        tox_ini = ToxIni(tox_ini_file)
        config = Config(tox_ini, overrides=[], root=tmp_path, pos_args=pos_args, work_dir=tmp_path)
        loader = config.get_env("py").loaders[0]
        return loader.load(key="env", of_type=str, conf=config, env_name="a")

    return example  # noqa



import json
import shutil
import sys

args = {
    "stdout": sys.stdout.isatty(),
    "stderr": sys.stderr.isatty(),
    "stdin": sys.stdin.isatty(),
    "terminal": shutil.get_terminal_size(fallback=(-1, -1)),
}
result = json.dumps(args)
print(result)



"""
Customize argparse logic for tox (also contains the base options).
"""

import argparse
import logging
import os
import sys
from argparse import SUPPRESS, Action, ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Type, TypeVar, cast

from tox.config.loader.str_convert import StrConvert
from tox.plugin import NAME
from tox.session.state import State

from .env_var import get_env_var
from .ini import IniConfig

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from typing import Literal
else:  # pragma: no cover (py38+)
    from typing_extensions import Literal


class ArgumentParserWithEnvAndConfig(ArgumentParser):
    """
    Argument parser which updates its defaults by checking the configuration files and environmental variables.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # sub-parsers also construct an instance of the parser, but they don't get their own file config, but inherit
        self.file_config = kwargs.pop("file_config") if "file_config" in kwargs else IniConfig()
        kwargs["epilog"] = self.file_config.epilog
        super().__init__(*args, **kwargs)

    def fix_defaults(self) -> None:
        for action in self._actions:
            self.fix_default(action)

    def fix_default(self, action: Action) -> None:
        if hasattr(action, "default") and hasattr(action, "dest") and action.default != SUPPRESS:
            of_type = self.get_type(action)
            key = action.dest
            outcome = get_env_var(key, of_type=of_type)
            if outcome is None and self.file_config:
                outcome = self.file_config.get(key, of_type=of_type)
            if outcome is not None:
                action.default, default_value = outcome
                action.default_source = default_value  # type: ignore[attr-defined]
        if isinstance(action, argparse._SubParsersAction):  # noqa
            for values in action.choices.values():  # noqa
                if not isinstance(values, ToxParser):  # pragma: no cover
                    raise RuntimeError("detected sub-parser added without using our own add command")
                values.fix_defaults()

    @staticmethod
    def get_type(action: Action) -> Type[Any]:
        of_type: Optional[Type[Any]] = getattr(action, "of_type", None)
        if of_type is None:
            if isinstance(action, argparse._AppendAction):  # noqa
                of_type = List[action.type]  # type: ignore[name-defined]
            elif isinstance(action, argparse._StoreAction) and action.choices:  # noqa
                loc = locals()
                loc["Literal"] = Literal
                as_literal = f"Literal[{', '.join(repr(i) for i in action.choices)}]"
                of_type = eval(as_literal, globals(), loc)
            elif action.default is not None:
                of_type = type(action.default)
            elif isinstance(action, argparse._StoreConstAction) and action.const is not None:  # noqa
                of_type = type(action.const)
            else:
                raise TypeError(action)
        return of_type


class HelpFormatter(ArgumentDefaultsHelpFormatter):
    """
    A help formatter that provides the default value and the source it comes from.
    """

    def __init__(self, prog: str) -> None:
        super().__init__(prog, max_help_position=30, width=240)

    def _get_help_string(self, action: Action) -> Optional[str]:

        text: str = super()._get_help_string(action) or ""  # noqa
        if hasattr(action, "default_source"):
            default = " (default: %(default)s)"
            if text.endswith(default):
                text = f"{text[: -len(default)]} (default: %(default)s -> from %(default_source)s)"
        return text


Handler = Callable[[State], int]

ToxParserT = TypeVar("ToxParserT", bound="ToxParser")
DEFAULT_VERBOSITY = 2


class Parsed(Namespace):
    @property
    def verbosity(self) -> int:
        result: int = max(self.verbose - self.quiet, 0)
        return result

    @property
    def is_colored(self) -> bool:
        return cast(bool, self.colored == "yes")


ArgumentArgs = Tuple[Tuple[str, ...], Optional[Type[Any]], Dict[str, Any]]


class ToxParser(ArgumentParserWithEnvAndConfig):
    """Argument parser for tox."""

    def __init__(self, *args: Any, root: bool = False, add_cmd: bool = False, **kwargs: Any) -> None:
        self.of_cmd: Optional[str] = None
        self.handlers: Dict[str, Tuple[Any, Handler]] = {}
        self._arguments: List[ArgumentArgs] = []
        self._groups: List[Tuple[Any, Dict[str, Any], List[Tuple[Dict[str, Any], List[ArgumentArgs]]]]] = []
        super().__init__(*args, **kwargs)
        if root is True:
            self._add_base_options()
        if add_cmd is True:
            msg = "tox command to execute (default to legacy if not specified)"
            self._cmd: Optional[Any] = self.add_subparsers(title="command", help=msg, dest="command")
            self._cmd.required = False
            self._cmd.default = "legacy"
        else:
            self._cmd = None

    def add_command(self, cmd: str, aliases: Sequence[str], help_msg: str, handler: Handler) -> "ArgumentParser":
        if self._cmd is None:
            raise RuntimeError("no sub-command group allowed")
        sub_parser: ToxParser = self._cmd.add_parser(
            cmd, help=help_msg, aliases=aliases, formatter_class=HelpFormatter, file_config=self.file_config
        )
        sub_parser.of_cmd = cmd  # mark it as parser for a sub-command
        content = sub_parser, handler
        self.handlers[cmd] = content
        for alias in aliases:
            self.handlers[alias] = content
        for (args, of_type, kwargs) in self._arguments:
            sub_parser.add_argument(*args, of_type=of_type, **kwargs)
        for (args, kwargs, excl) in self._groups:
            group = sub_parser.add_argument_group(*args, **kwargs)
            for (e_kwargs, arguments) in excl:
                excl_group = group.add_mutually_exclusive_group(**e_kwargs)
                for (a_args, _, a_kwargs) in arguments:
                    excl_group.add_argument(*a_args, **a_kwargs)
        return sub_parser

    def add_argument_group(self, *args: Any, **kwargs: Any) -> Any:
        result = super().add_argument_group(*args, **kwargs)
        if self.of_cmd is None:
            if args not in (("positional arguments",), ("optional arguments",)):

                def add_mutually_exclusive_group(**e_kwargs: Any) -> Any:
                    def add_argument(*a_args: str, of_type: Optional[Type[Any]] = None, **a_kwargs: Any) -> Action:
                        res_args: Action = prev_add_arg(*a_args, **a_kwargs)  # type: ignore[has-type]
                        arguments.append((a_args, of_type, a_kwargs))
                        return res_args

                    arguments: List[ArgumentArgs] = []
                    excl.append((e_kwargs, arguments))
                    res_excl = prev_excl(**kwargs)
                    prev_add_arg = res_excl.add_argument
                    res_excl.add_argument = add_argument  # type: ignore[assignment]
                    return res_excl

                prev_excl = result.add_mutually_exclusive_group
                result.add_mutually_exclusive_group = add_mutually_exclusive_group  # type: ignore[assignment]
                excl: List[Tuple[Dict[str, Any], List[ArgumentArgs]]] = []
                self._groups.append((args, kwargs, excl))
        return result

    def add_argument(self, *args: str, of_type: Optional[Type[Any]] = None, **kwargs: Any) -> Action:
        result = super().add_argument(*args, **kwargs)
        if self.of_cmd is None and (result.dest not in ("help",)):
            self._arguments.append((args, of_type, kwargs))
            if hasattr(self, "_cmd") and self._cmd is not None and hasattr(self._cmd, "choices"):
                for parser in {id(v): v for k, v in self._cmd.choices.items()}.values():
                    parser.add_argument(*args, of_type=of_type, **kwargs)
        if of_type is not None:
            result.of_type = of_type  # type: ignore[attr-defined]
        return result

    @classmethod
    def base(cls: Type[ToxParserT]) -> ToxParserT:
        return cls(add_help=False, root=True)

    @classmethod
    def core(cls: Type[ToxParserT]) -> ToxParserT:
        return cls(prog=NAME, formatter_class=HelpFormatter, add_cmd=True, root=True)

    def _add_base_options(self) -> None:
        """Argument options that always make sense."""
        add_core_arguments(self)
        self.fix_defaults()

    def parse_known_args(  # type: ignore[override]
        self, args: Optional[Sequence[str]], namespace: Optional[Parsed] = None
    ) -> Tuple[Parsed, List[str]]:
        if args is None:
            args = sys.argv[1:]
        cmd_at: Optional[int] = None
        if self._cmd is not None and args:
            for at, arg in enumerate(args):
                if arg in self._cmd.choices:
                    cmd_at = at
                    break
            else:
                cmd_at = None
        if cmd_at is not None:  # if we found a command move it to the start
            args = args[cmd_at], *args[:cmd_at], *args[cmd_at + 1 :]
        elif args not in (("--help",), ("-h",)) and (self._cmd is not None and "legacy" in self._cmd.choices):
            # on help no mangling needed, and we also want to insert once we have legacy to insert
            args = "legacy", *args
        result = Parsed() if namespace is None else namespace
        _, args = super().parse_known_args(args, namespace=result)
        return result, args


def add_verbosity_flags(parser: ArgumentParser) -> None:
    from tox.report import LEVELS

    level_map = "|".join("{} - {}".format(c, logging.getLevelName(l)) for c, l in sorted(list(LEVELS.items())))
    verbosity_group = parser.add_argument_group(
        f"verbosity=verbose-quiet, default {logging.getLevelName(LEVELS[3])}, map {level_map}",
    )
    verbosity = verbosity_group.add_mutually_exclusive_group()
    verbosity.add_argument(
        "-v", "--verbose", action="count", dest="verbose", help="increase verbosity", default=DEFAULT_VERBOSITY
    )
    verbosity.add_argument("-q", "--quiet", action="count", dest="quiet", help="decrease verbosity", default=0)


def add_color_flags(parser: ArgumentParser) -> None:
    converter = StrConvert()
    if converter.to_bool(os.environ.get("NO_COLOR", "")):
        color = "no"
    elif converter.to_bool(os.environ.get("FORCE_COLOR", "")):
        color = "yes"
    else:
        color = "yes" if sys.stdout.isatty() else "no"

    parser.add_argument(
        "--colored",
        default=color,
        choices=["yes", "no"],
        help="should output be enriched with colors",
    )


def add_core_arguments(parser: ArgumentParser) -> None:
    add_color_flags(parser)
    add_verbosity_flags(parser)
    parser.add_argument(
        "-c",
        "--conf",
        dest="config_file",
        metavar="file",
        default=None,
        type=Path,
        of_type=Optional[Path],
        help="configuration file for tox (if not specified will discover one)",
    )
    parser.add_argument(
        "--workdir",
        dest="work_dir",
        metavar="dir",
        default=None,
        type=Path,
        of_type=Optional[Path],
        help="tox working directory (if not specified will be the folder of the config file)",
    )
    parser.add_argument(
        "--root",
        dest="root_dir",
        metavar="dir",
        default=None,
        type=Path,
        of_type=Optional[Path],
        help="project root directory (if not specified will be the folder of the config file)",
    )


__all__ = (
    "DEFAULT_VERBOSITY",
    "Parsed",
    "ToxParser",
    "Handler",
)



from pathlib import Path

from packaging.requirements import Requirement
from pytest_mock import MockerFixture

from tox.pytest import MonkeyPatch, ToxProjectCreator
from tox.tox_env.python.api import PythonDep


def test_deps_path_eq() -> None:
    dep_1 = PythonDep(Path.cwd())
    dep_2 = PythonDep(Path.cwd())
    assert dep_1 == dep_2


def test_deps_path_ne() -> None:
    dep_1 = PythonDep(Path.cwd())
    dep_2 = PythonDep(Path.cwd() / "a")
    assert dep_1 != dep_2


def test_deps_req_eq() -> None:
    dep_1 = PythonDep(Requirement("pytest"))
    dep_2 = PythonDep(Requirement("pytest"))
    assert dep_1 == dep_2


def test_deps_req_ne() -> None:
    dep_1 = PythonDep(Requirement("pytest"))
    dep_2 = PythonDep(Requirement("tox"))
    assert dep_1 != dep_2


def test_deps_repr() -> None:
    dep_1 = PythonDep(Path.cwd())
    assert repr(dep_1) == f"PythonDep(value={Path.cwd()!r})"


def test_requirements_txt(tox_project: ToxProjectCreator, monkeypatch: MonkeyPatch, mocker: MockerFixture) -> None:
    prj = tox_project(
        {
            "tox.ini": "[testenv]\npackage=skip\ndeps=-rrequirements.txt",
            "requirements.txt": "nose",
        }
    )
    execute_calls = prj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = prj.run("r", "-e", "py")
    result.assert_success()
    tox_env = result.state.tox_env("py")

    assert execute_calls.call_count == 1
    exp = [str(tox_env.conf["env_python"]), "-I", "-m", "pip", "install", "-r"]
    got_cmd = execute_calls.call_args[0][2].cmd

    assert got_cmd[:-1] == exp

    req = Path(got_cmd[-1])
    assert req.parent == tox_env.core["tox_root"]
    assert req.name.startswith("requirements-")
    assert req.name.endswith(".txt")
